; Copyright (c) 2007, Vania Da Deppo, Alice Dalla Stella, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_level2_to_level2_nac, header_in, image_in, header_out, image_out, quality_map, sigma_map
;
; PURPOSE:
;   Main interface function to the OSIRIS level 2 to level 3 image converter
;   The function will convert from calibrated [W/m^2/sr/nm] unit images to images where the geometric distortion has been removed
;   The function only works for nac images
;
; CATEGORY:
;   Osiris Geometric Calibration (nac)
;
; CALLING SEQUENCE:
;  Result = osiris_level2_to_level3_nac(header_in, image_in, header_out, image_out, quality_map, sigma_map)
;
; INPUTS:
;   header_in:   	input level 1 PDS header structure
;   image_in:  		input level 1 image data
; 	quality_map		The quality map of the image
;	sigma_map:		The sigma map of the image
;
; KEYWORD PARAMETERS:
;   CODMAC:     If set the CODMAC data levels are used - else OSIRIS
;
; OUTPUTS:
;   header_out:     output PDS image header
;   image_out:       output calibrated image data
; 	quality_map		The quality map of the image
;	sigma_map:		The sigma map of the image
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
; EXAMPLE:
;   p_read, pickfile(), h_in, d_in                 ; read in raw PDS image
;   print, osiris_level1_to_level2(h_in, d_in, h_out2, d_out2) ; calibrate the image
;   print, osiris_level2_to_level3_nac(h_out2, d_out2, h_out3, d_out3) ; calibrate the image
;   p_write, pickfile(), h_out3, d_out3             ; save the image to a new PDS file
;
; MODIFICATION HISTORY:
;   Written:  01/03/2007  Vania Da Deppo, Alice Dalla Stella
;   Modified: 2008-09-08  Stubbe F. Hviid  - to fit calibration pipeline conventions
;-
FUNCTION osiris_level2_to_level3_nac,header_in, image_in, header_out, image_out, quality_map, sigma_map, DEBUG=GEBUG
VERSION = '1.0'

    ; initiate the output header
    image_out = double(image_in)
    header_out= header_in


     ; Verify the the correction has not already been performed
    flag = p_value(header_out, 'LEVEL3_PIPELINE.GEOMETRIC_CORRECTION_FLAG', /HISTORY)

    if flag eq 'TRUE' then begin
       if keyword_set(VERBOSE) then print, 'osiris_Correction skipped - Already performed'
       return, 1
    endif

    ; Extract model information
    osiris_model, header_out, camera, model

    ; check for valid camera
    if camera ne 'NAC' then begin
        print, 'Illegal camera for function "osiris_level2_to_level3_nac" - ' + camera
        return, 0
    endif

    ;
    ; INITIALIZATION
    ; --------------
    ; default values and initialization
    ;
    debug     = keyword_set(debug)

    if (debug) then print,'M: LEVEL2_TO_LEVEL3_NAC_02 - BEGIN'
    if n_elements(method) eq 0 then method = 'BL'

    ;
    ; extract some header information
    bin       = p_value(header_out, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
    x_start   = p_value(header_out, 'IMAGE.FIRST_LINE_SAMPLE')-1
    y_start   = p_value(header_out, 'IMAGE.FIRST_LINE')-1
    swbin     = (p_value(header_out, 'SR_COMPRESSION.PIXEL_AVERAGING_WIDTH'))[0]

    ; extract the effective binning size
    bf = 1
    if bin eq '2x2' then bf = 2
    if bin eq '4x4' then bf = 4
    if bin eq '8x8' then bf = 8

    if swbin gt 1 then bf = bf * swbin

    bin = bf


    ;
    ; load boresight information
    ; ..........................
    ; boresight data file
    ;
    fBoresight = p_resolve_database_file('DISTORSION', camera, model, 'BORE', '', 'TXT')
    if fBoresight eq '' then begin
        print, 'ERROR - Cannot load NAC boresight data from the database'
        return, 0
    endif
    ; load the boresight data
    bor = init_nac_boresight(fBoresight, IFAIL=ifail, DEBUG=debug)
    if ifail ne 0 then begin
        print, 'ERROR - Format error in boresight data file'
        return, 0
    endif

    ;
    ; load images containing de-distorted pixel coordinates
    ; .....................................................
    fDX = p_resolve_database_file('DISTORSION', camera, model, 'BIN'+strtrim(bin, 2) + '-X', '', '')
    if fDX eq '' then begin
        print, 'ERROR - cannot find X distosion data'
        return, 0
    endif

    fDY = p_resolve_database_file('DISTORSION', camera, model, 'BIN'+strtrim(bin, 2) + '-Y', '', '')
    if fDY eq '' then begin
        print, 'ERROR - cannot find Y distosion data'
        return, 0
    endif

    ; read distorsion coordinates
    p_read, fDX, hdx, ddx
    p_read, fDY, hdy, ddy

    ;
    ; extract information on image geometry from 'geomdef' structure
    ;
    wh = size(image_out, /DIM)
    binx = bin
    biny = bin
    x1   = x_start
    y1   = y_start
    x2   = x1 + wh[0] * (2^bin)
    y2   = y1 + wh[1] * (2^bin)
    nx   = wh[0]
    ny   = wh[1]

    if (debug) then begin
        print,'M:Binning factor: ',binx,'x',biny,'.'
        print,'M:Window        : [',x1,':',x2,',',y1,':',y2,'] (unbinned).'
        print,'M:Image size    : ',nx,'x',ny,' pixels (binned).'
    endif

    ;
    ; LIMITS OF WINDOW IN PIXELS OF THE IMAGE IN WHICH PIXEL COORDINATES ARE STORED
    ; -----------------------------------------------------------------------------
    ; compute limits of window in unbinned coordinates (RCCD frame)
    ;
    if (debug) then print,'M: -- WINDOW DEFINITION:'

    x1u = float(binx)/2.0+float(x1)-0.5
    y1u = float(biny)/2.0+float(y1)-0.5
    x2u = x1u+float((nx-1)*binx)
    y2u = y1u+float((ny-1)*biny)

    if (debug) then print,'M:Window        : [',x1u,':',x2u,',',y1u,':',y2u,'] (RCCD).'

    ;
    ; derive X,Y limits of window in coordinates images
    ;
    bin_test_x = binx mod 2       ; is the X binning factor odd/even?
    bin_test_y = biny mod 2       ; is the Y binning factor odd/even?
    feps       = 1.0E-2

    t = bin_test_x+bin_test_y

    if (t ne 0) and (t ne 2) then begin
        print,'error - Binning in X and Y are incompatible!'
        return, 0
    endif

    if (bin_test_x eq 1) then begin        ; X -- ODD/IMPAIR
        x1ui = fix(x1u+feps)
        x2ui = fix(x2u+feps)
    endif else begin                       ; X -- EVEN/PAIR
        x1ui = fix(x1u-0.5+feps)
        x2ui = fix(x2u-0.5+feps)
    endelse

    if (bin_test_y eq 1) then begin        ; Y -- ODD/IMPAIR
        y1ui = fix(y1u+feps)
        y2ui = fix(y2u+feps)
    endif else begin                       ; Y -- EVEN/PAIR
        y1ui = fix(y1u-0.5+feps)
        y2ui = fix(y2u-0.5+feps)
    endelse

    ;
    ; size of sub-image
    ;
    px    = (x2ui-x1ui)/binx+1
    py    = (y2ui-y1ui)/biny+1
    ppix  = long(px)*long(py)

    if (debug) then begin
        print,'M:Window          : [',x1ui,':',x2ui,',',y1ui,':',y2ui,'] (ICOR).'
        print,'M:Step            : [',binx,',',biny,'].'
        print,'M:Size            : ',px,'x',py,' pixels.'
        print,'M:Number of pixels: ',ppix,'.'
    endif

    ;
    ; EXTRACT PIXEL COORDINATES AND BUILD LIST OF PIXEL COORDINATES
    ; -------------------------------------------------------------
    if (debug) then print,'M: -- LIST OF PIXEL COORDINATES:'

    imax = ddx
    imay = ddy

    ;
    ; create 'xarr' and 'yarr' variables
    ;
    xarr = fltarr(px,py)
    yarr = fltarr(px,py)
    dind = lonarr(ppix)

    ;
    ; build arrays containing indices of pixels of coordinate images (1D/long)
    ;
    if (debug) then print,'M:Building array containing indices of image pixels.'

    nxima = 2048
    nyima = 2048
    plist = indgen(px)*binx+x1ui-1

    for j=0,py-1 do begin
        iy        = (j*biny+y1ui-1) > 0
        xarr(*,j) = imax(plist,iy)-x1u+1.0
        yarr(*,j) = imay(plist,iy)-y1u+1.0
    endfor

    ;
    ; normal end of routine
    ;
    if (debug) then print,'M: COMPUTE_PIXEL_COORDINATES_02 - END'

    ;
    ; resamples input image and creates output image 'kim'
    ;
    if (debug) then print,'M:Resampling image.'

    kim = fltarr(nx,ny)
    resample_nac_image, xarr, yarr, double(image_in), image_out, METHOD=method, DEBUG=debug, IFAIL=ifail

	if n_elements(quality_map) ne 0 then begin
		resample_nac_image, xarr, yarr, double(quality_map), quality_map_out, METHOD=method, DEBUG=debug, IFAIL=ifail
		quality_map = float(quality_map_out)
	endif

	if n_elements(sigma_map) ne 0 then begin
		resample_nac_image, xarr, yarr, double(sigma_map), sigma_map_out, METHOD=method, DEBUG=debug, IFAIL=ifail
		sigma_map = float(sigma_map_out)
	endif

    if (ifail ne 0) then begin
        print,'ERROR - Error calling RESAMPLE_IMAGE_02 routine!'
        return, 0
    endif

    geom = EXTRACT_IMAGE_GEOMETRY_NAC(header_in, DEBUG=debug, IFAIL=ifail, LOS=los_st)
    if ifail ne 0 then begin
        print, 'ERROR: Error extracting geometry information'
        return, 0
    endif

    ;
    ; compute LOS of reference pixel and camera ROLL
    ;
    los_rp = compute_nac_boresight(los_st, bor, DEBUG=debug, IFAIL=ifail)

    if (ifail ne 0) then begin
        print,'ERROR - Error calling COMPUTE_BORESIGHT_01 routine!'
        return, 0
    endif

    ; Update level 3 header
    if (debug) then print,'> Updating PDS header -> level 3.'
    update_level3_header, header_out, bor, los_rp, method

    ; update header
    datafiles = p_value(header_out, 'LEVEL3_PIPELINE.CALIBRATION_DATAFILES', /HISTORY)
    fsb = p_parse_file_path(fBoresight)
    fsdx = p_parse_file_path(fDX)
    fsdy = p_parse_file_path(fDY)
    datafiles = [datafiles, fsdx.filename, fsdy.filename, fsb.filename]
    p_setvalue, header_out, 'LEVEL3_PIPELINE.CALIBRATION_DATAFILES', datafiles, /HISTORY

    ; Change the processing flag
    p_setvalue, header_out, 'LEVEL3_PIPELINE.GEOMETRIC_CORRECTION_FLAG', 'TRUE', /LABEL, /NO_QUOTES, /HISTORY
    p_setvalue, header_out, 'LEVEL3_PIPELINE.GEOMETRIC_CORRECTION_VERSION', VERSION, /GROUP, /HISTORY

    print, 'SUCCESS - Geometric distorsion correction'
    return, 1
end