; Copyright (c) 2007, Vania Da Deppo, Alice Dalla Stella, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_level2_to_level2_wac, header_in, image_in, header_out, image_out, quality_map, sigma_map
;
; PURPOSE:
;   Main interface function to the OSIRIS level 2 to level 3 image converter
;   The function will convert from calibrated [W/m^2/sr/nm] unit images to images where the geometric distortion has been removed
;   The function only works for WAC images
;
; CATEGORY:
;   Osiris Geometric Calibration (WAC)
;
; CALLING SEQUENCE:
;  Result = osiris_level2_to_level3_wac(header_in, image_in, header_out, image_out, quality_map, sigma_map)
;
; INPUTS:
;   header_in:   input level 1 PDS header structure
;   image_in:  input level 1 image data
; 	quality_map		The quality map of the image
;	sigma_map:		The sigma map of the image
;
; KEYWORD PARAMETERS:
;   CODMAC:     If set the CODMAC data levels are used - else OSIRIS
;
; OUTPUTS:
;   header_out:     output PDS image header
;   image_out:       output calibrated image data
; 	quality_map		The quality map of the image
;	sigma_map:		The sigma map of the image
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
; EXAMPLE:
;   p_read, pickfile(), h_in, d_in                 ; read in raw PDS image
;   print, osiris_level1_to_level2(h_in, d_in, h_out2, d_out2) ; calibrate the image
;   print, osiris_level2_to_level3_wac(h_out2, d_out2, h_out3, d_out3) ; calibrate the image
;   p_write, pickfile(), h_out3, d_out3             ; save the image to a new PDS file
;
; MODIFICATION HISTORY:
;   Written:  01/03/2007  Vania Da Deppo, Alice Dalla Stella
;   Modified: 2008-09-08  Stubbe F. Hviid  - to fit calibration pipeline conventions
;-
FUNCTION osiris_level2_to_level3_wac, header_in, image_in, header_out, image_out, quality_map, sigma_map
VERSION = '1.0'

    ; initiate the output header
    image_out = double(image_in)
    header_out= header_in

     ; Verify the the correction has not already been performed
    flag = p_value(header_out, 'LEVEL3_PIPELINE.GEOMETRIC_CORRECTION_FLAG', /HISTORY)

    if flag eq 'TRUE' then begin
       if keyword_set(VERBOSE) then print, 'osiris_Correction skipped - Already performed'
       return, 1
    endif

    ; Extract model information
    osiris_model, header_out, camera, model

    ; check for valid camera
    if camera ne 'WAC' then begin
        print, 'Illegal camera for function "osiris_level2_to_level3_wac" - ' + camera
        return, 0
    endif

    ; extract some header information
    bin       = p_value(header_out, 'SR_ACQUIRE_OPTIONS.HARDWARE_BINNING_ID')
    x_start   = p_value(header_out, 'IMAGE.FIRST_LINE_SAMPLE')-1
    y_start   = p_value(header_out, 'IMAGE.FIRST_LINE')-1
    swbin     = (p_value(header_out, 'SR_COMPRESSION.PIXEL_AVERAGING_WIDTH'))[0]

    ; extract the effective binning size
    bf = 1
    if bin eq '2x2' then bf = 2
    if bin eq '4x4' then bf = 4
    if bin eq '8x8' then bf = 8

    if swbin gt 1 then bf = bf * swbin

    bin = bf

    ; extract image size
    dim = size(image_out, /DIM)

    ; handle windowing
    A_tot=dblarr(2048/bin,2048/bin)

    if bin gt 1 then begin
        A_tmp = image_out
        x_start_b = round(x_start/bin)
        y_start_b = round(y_start/bin)
        A_tot[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1] = A_tmp
    endif else begin
        A_tot[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1] = image_out
    endelse

    ; look up distortion partameter file
    fKX = p_resolve_database_file('DISTORSION', camera, model, 'KX' + strtrim(bin, 2), '', 'TXT')
    if fKX eq '' then begin
        print, 'ERROR: cannot resolve KX distotion map'
        return, 0
    endif

    fKY = p_resolve_database_file('DISTORSION', camera, model, 'KY' + strtrim(bin, 2), '', 'TXT')
    if fKY eq '' then begin
        print, 'ERROR: cannot resolve KY distotion map'
        return, 0
    endif

    ; read the distortion parameters
	KX_txt = p_read_text_file(fKX, /STRIP_PDS_HEADER);
	KX=dblarr(4,4)
	for i=0, 3 do begin
		KX[*, i] = double(strsplit(KX_txt[i], ' ', /EXTRACT))
	endfor

	KY_txt = p_read_text_file(fKY, /STRIP_PDS_HEADER);
	KY=dblarr(4,4)
	for i=0, 3 do begin
		KY[*, i] = double(strsplit(KY_txt[i], ' ', /EXTRACT))
	endfor

    ; do processing
    size_x=2080/bin
    size_y=2080/bin

    ; Correct the image based on Kx and Ky with POLY_2D:
    m = max(A_tot)
    out_tmp = POLY_2D(A_tot, KX, KY,1,size_x,size_y,cubic=-0.1,missing=m)

    ; extract the image data back out of the working buffer
    if bin gt 1 then begin
        image_out = out_tmp[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1]
    endif else begin
        image_out = out_tmp[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1]
    endelse

    ; extract the image data back out of the working buffer
    if bin gt 1 then begin
        image_out = out_tmp[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1]
    endif else begin
        image_out = out_tmp[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1]
    endelse

	; handle the quality_map (if existing)
	if n_elements(quality_map) ne 0 then begin
		; work around for a header error existing on a few images
		sq = size(quality_map, /dim)
		if sq[0] gt dim[0] OR sq[1] gt dim[1] then quality_map = quality_map[0:dim[0]-1, 0:dim[1]-1]
		; end work around

		if bin gt 1 then begin
        	A_tot[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1] = quality_map
    	endif else begin
        	A_tot[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1] = quality_map
    	endelse
    	m = max(A_tot)
    	out_tmp = POLY_2D(A_tot, KX, KY,1,size_x,size_y,cubic=-0.1,missing=m)
    	if bin gt 1 then begin
        	quality_map = out_tmp[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1]
    	endif else begin
        	quality_map = out_tmp[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1]
    	endelse
    endif

	; handle the quality_map (if existing)
	if n_elements(sigma_map) ne 0 then begin
		; work around for a header error existing on a few images
		ss = size(sigma_map, /dim)
		if ss[0] gt dim[0] OR ss[1] gt dim[1] then sigma_map = sigma_map[0:dim[0]-1, 0:dim[1]-1]
		; end work around

		if bin gt 1 then begin
        	A_tot[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1] = sigma_map
    	endif else begin
        	A_tot[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1] = sigma_map
    	endelse
    	m = max(A_tot)
    	out_tmp = POLY_2D(A_tot, KX, KY,1,size_x,size_y,cubic=-0.1,missing=m)
    	if bin gt 1 then begin
        	sigma_map = out_tmp[x_start_b:x_start_b+dim[0]-1,y_start_b:y_start_b+dim[1]-1]
    	endif else begin
        	sigma_map = out_tmp[x_start:x_start+dim[0]-1,y_start:y_start+dim[1]-1]
    	endelse
    endif


    ; update header
    datafiles = p_value(header_out, 'LEVEL3_PIPELINE.CALIBRATION_DATAFILES', /HISTORY)
    fsx = p_parse_file_path(fKX)
    fsy = p_parse_file_path(fKY)
    datafiles = [datafiles, fsx.filename, fsy.filename]
    p_setvalue, header_out, 'LEVEL3_PIPELINE.CALIBRATION_DATAFILES', datafiles, /HISTORY

    ; Change the processing flag
    p_setvalue, header_out, 'LEVEL3_PIPELINE.GEOMETRIC_CORRECTION_FLAG', 'TRUE', /LABEL, /NO_QUOTES, /HISTORY
    p_setvalue, header_out, 'LEVEL3_PIPELINE.GEOMETRIC_CORRECTION_VERSION', VERSION, /GROUP, /HISTORY


    print, 'SUCCESS - Geometric distorsion correction'
    return, 1
end