; RMV_CHN_NOI.PRO  by  R. Kramm  23-JUL-2003, 04-NOV-2004
; =======================================================
;
;+
; NAME:
;   FUNCTION osiris_rmv_chn_noi, header, image
;
; PURPOSE:
;   Remove coherent noise from image
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_rmv_chn_noi(header, image)
;
; INPUTS:
;   header:   The PDS image header of the image to be corrected
;   image:      The image data of the image to be corrected
;
; KEYWORD PARAMETERS:
;   NONE
;
; OUTPUTS:
;   amp:		The amplitude of the CHN in x and y
;   anz:		?
;
; RETURNS:
;   image with same dimensions as the input image containing the noise component
;   On error -1
;
; RESTRICTIONS:
;	None
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   noise = osiris_rmv_chn_noi(h, d, amp, anz)
;   d = d - noise
;
; MODIFICATION HISTORY:
;   Written by:  			R. Kramm   23-JUL-2003, 04-NOV-2004
;   Made PDS compatible		S.F. Hviid 2-JUL-2005
;-

FUNCTION osiris_rmv_chn_noi, H, P, AMP, ANZ, quiet_=quiet_
;
;  Function that determines the coherent noise components (Amp > 0.58 DU).
;  Coh_Noise = RMV_CHN_NOI(Header, Pixel);  Result is the extracted noise.
;  In case of split readout (AB), both sides will be processed separately.
; ------------------------------------------------------------------------
;
	AMP = fltarr(2)
	ANZ = intarr(2)
;
; ------------------------------------------------------------------------
;
	LIM = 0.58	;67                            ; max  coherent noise amplitude

; Extract the dimensions of the image
	SZE = SIZE(P)
	DX = SZE(1)
	DY = SZE(2)

; declare output image array
	CH = dblarr(DX, DY)
	CH[*,*] = 0.0

;
; Determine the amplifier channel
	CHL = p_value(header, 'SR_ACQUIRE_OPTIONS.AMPLIFIER_ID')

	; amplifier is BOTH we need to handle tyhe two sides of the image seperately
	block_count = 1
	IF (CHL EQ 'BOTH')  THEN  begin
		DX = DX/2L
		block_count = 2
	ENDIF
;
; ------------------------------------------------------------------------
;

; Ensure that we have enough data to do the noise correction

	; Exit on mini image
	IF (DX<DY LT 16)  THEN  RETURN, CH
;
	; Check for image dimensions
	IF (((DX AND 3) > (DY AND 1)) NE 0)  THEN  BEGIN
		PRINT, ' '
		PRINT, ' RMV_CHN_NOI:  Warning.   Image size not (N*4 x M*2) pixel.'
		PRINT, ' Current size not an optimum for expected noise components.'
		PRINT, ' '
	ENDIF
;
; ------------------------------------------------------------------------
;

; Call the repair subroutine
	; osiris_repair, H, P, HH, PP, /FILTER, /QUIET                ; ToDo:  Reenable the repair function
	HH = h
	PP = P

;
FOR K = 0,block_count-1 DO BEGIN
	XRG = [0L, DX-1L] + K*DX				; half size on split readout
	BLX = LONG(0.05*DX) > 3L				; low x frequencies discarded
	BLY = LONG(0.05*DY) > 3L				; low y frequencies discarded
    ;
    FT  = FFT(FLOAT(osiris_his_clamp(PP(XRG(0):XRG(1),*))), -1)
    ;
    PWR = ABS(FT)				&  PWR = PWR-(SHIFT(PWR,-1,0)<SHIFT(PWR,1,0))
	PWR = SHIFT(PWR, BLX, BLY)  &  PWR(0,0) = COMPLEXARR(2L*BLX+1L,2L*BLY+1L)
	PWR = SHIFT(PWR,-BLX,-BLY)  &  PWR(0,0) = COMPLEXARR(1,DY)
	;
	POS = WHERE(PWR GE LIM)
    ;
    ; --------------------------------------------------------------------
    ;											too many frequencies found
    IF (N_ELEMENTS(POS) GT 33)  THEN  BEGIN
		POS = [DX/4L*[1L,2L,3L], (DX*DY/2L + DX/4L*[1L,2L,3L])]
		ARR = ABS(FT(POS)) - (ABS(FT(POS-1L)) > ABS(FT(POS+1L)))
		POS = POS(WHERE(ARR GE LIM, CNT) > 0L)
		IF (CNT EQ 0)  THEN  POS = -1
    ENDIF
    ;
    ; --------------------------------------------------------------------
    ;
    IF (TOTAL(POS) GT 0L)  THEN  BEGIN
       IF (POS(0) EQ 0L)  THEN  POS = POS(1:*)
       ;
       COH = COMPLEXARR(DX,DY)
       COH(POS) = FT(POS)
       COH = FLOAT(FFT(COH, 1, /DOUBLE))
    ENDIF  ELSE  COH = FLTARR(DX,DY)
    ;
    ; --------------------------------------------------------------------
    ;
    AMP(K) = COH(0) / (ABS(COH(0)) > 1.0E-30) * MEAN(ABS(COH))
    ;
    POS = POS(WHERE(POS LE (DX*DY/2L + DX/2L)))
    POS = POS(REVERSE(SORT(ABS(FT(POS)))))
    LEN = STRTRIM(N_ELEMENTS(POS),2)
    ;
    ANZ(K) = FIX(N_ELEMENTS(POS))
    ;
	IF NOT KEYWORD_SET(quiet_)  THEN  BEGIN
    	IF (ANZ(K) LE 5)  THEN  TXT = ''  ELSE    $
        	TXT = STRING('ANZ = ',STRTRIM(ANZ(K),2),'(A16,A)')
		;
		PRINT, ' '
		PRINT, ' Pos x ='+ STRING(FIX(POS MOD DX), '('+LEN+'I8)')
		PRINT, ' Pos y ='+ STRING(FIX(POS  /  DX), '('+LEN+'I8)')
		PRINT, ' Ampl. ='+ STRING(PWR(POS),    '('+LEN+'F8.1)')
		PRINT, ' Total ='+ STRING(AMP(K),'DU', '(F8.1,A3)') + TXT
	ENDIF
    ;
    ; --------------------------------------------------------------------
    ;
    CH(K*DX,0) = COH
    ;
ENDFOR
;
IF (CHL EQ 'B')  THEN  AMP = SHIFT(AMP,1)
IF (CHL EQ 'B')  THEN  ANZ = SHIFT(ANZ,1)
;
IF NOT KEYWORD_SET(quiet_)  THEN  PRINT, ' '
;
; ------------------------------------------------------------------------
;
; STOP, ' STOP @ END'
;
RETURN, CH
END
