; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_calc_poisson, header, image, quality_map, sigma_map
;
; PURPOSE:
;   Generate the primary error map from the poissong statistics of the image data
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_calc_poisson(header, image, quality_map, sigma_map)
;
; INPUTS:
;   header:   		The PDS image header of the image to be corrected
;   image:     		The image data of the image to be corrected
;   quality_map:	Existing map of quality estimate for each pixel to be modified
;   sigma_map:		Existing map of estimaped relative error for each pixel (in %) to be modified
;
; OUTPUTS:
;   sigma_map:		output map of estimaped relative error for each pixel (in %)
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
;   The function depends on an external data file <camera>_<unit>_BIAS_V<version>_<timestamp>.dat
;   to exist in the BIAS folder of the database tree. This file must have a tag_file compatible structure
;   The file must contain tags with names of the form:
;                                   <frame mode>_<amblifier>_<binning>
;   As a minimum the file must contain the tag "default" giving the default bias level to use if no better
;   value can be found
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   print, osiris_calc_poisson(h, d, q, s)
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   4/5-2009
;-

FUNCTION osiris_calc_poisson, header, image, quality_map, sigma_map
VERSION = '1.0'

	; first build the primary sigma_map
	gain = p_value(header, 'SR_ACQUIRE_OPTIONS.GAIN_ID')
	if gain eq 'HIGH' then begin
		tmp = image * 3.1
		sigma_map = sqrt(tmp) * 100.0 / tmp
	endif else begin
		tmp = image * 15.5
		sigma_map = sqrt(tmp) * 100.0 / tmp
	endelse


	p_setvalue, header, 'LEVEL2_PIPELINE.POISSON_NOISE_VERSION', VERSION, /GROUP, /HISTORY

	return, 1
END