; Copyright (c) 2004-2010, Stubbe F. Hviid
;   All rights reserved
;+
; NAME:
;   FUNCTION osiris_calibrate_noise, header, image, quality_map, sigma_map
;
; PURPOSE:
;   Remove coherent noise from image
;
; CATEGORY:
;   Osiris CCD Calibration
;
; CALLING SEQUENCE:
;  Result = osiris_calibrate_noise(header, image, quality_map_out, sigma_map_out)
;
; INPUTS:
;   header:   The PDS image header of the image to be corrected
;   image:      The image data of the image to be corrected
;   quality_map:	Existing map of quality estimate for each pixel to be modified
;   sigma_map:		Existing map of estimaped relative error for each pixel (in %) to be modified
;
; KEYWORD PARAMETERS:
;   NONE
;
; OUTPUTS:
;   header:   The PDS image header of the image after correction
;   image:      The image data of the image after correction
;   quality_map:	output map of quality estimate for each pixel
;   sigma_map:		output map of estimaped relative error for each pixel (in %)
;
; RETURNS:
;   0 on failure   1 on success
;
; RESTRICTIONS:
;   function requires that the paths defined in get_osiris_environment.pro has been set to correctly
;   point to the root directory of the external calibration files
;
;
; EXAMPLE:
;   p_read, pickfile(), h, d             ; read in raw PDS image
;   osiris_calibrate_noise, h, d
;
; MODIFICATION HISTORY:
;   Written by:  Stubbe F. Hviid   12/8-2004 based on p_adc_hi.pro by R. Kramm dated 29-OCT-2003
;-

FUNCTION osiris_calibrate_noise, header, image, quality_map, sigma_map, PARAMETER_FILELIST=PARAMETER_FILELIST, _EXTRA=_EXTRA
VERSION='V1.0'

    if keyword_set(VERBOSE) then print, 'osiris_calibrate_noise running'

    ; Verify the the correction has not already been performed
    flag = p_value(header, 'LEVEL2_PIPELINE.ROSETTA:COHERENT_NOISE_CORRECTION_FLAG', /HISTORY)

    if flag eq 'TRUE' then begin
       if keyword_set(VERBOSE) then print, 'osiris_calibrate_noise skipped - Already performed'
       return, 1
    endif

	; do the processing

	; first check if a CHN correction is required
	if mean(image) lt 5000 then begin
		image = image - osiris_rmv_chn_noi(header, image, amp, anz)
		proc_msg = 'Coherent Noise removed from image - amp = ' + strtrim(amp[0], 2) + '   anz = ' + strtrim(anz[0], 2)
	endif else proc_msg = 'No Coherent Noise removal performed - Not required'


    ; Change the processing flag
    p_setvalue, header, 'LEVEL2_PIPELINE.COHERENT_NOISE_CORRECTION_FLAG', 'TRUE', /LABEL, /NO_QUOTES, /HISTORY
    p_setvalue, header, 'LEVEL2_PIPELINE.COHERENT_NOISE_CORRECTION_VERSION', VERSION, /GROUP, /HISTORY


    ; finish
    if keyword_set(VERBOSE) then print, 'osiris_calibrate_noise - Done'

    return, 1
END